
#include <stdlib.h>
#include <stdio.h>
#include "f2c.h"
#include "fmmbase.P"
#include "fmmbase-util.h"
#include "fmmbase-io.h"

#include "fmm.h"

UL ll_type = 0;
L op_hi(void) {return wrap_hi("fmm V1");}
L op_libnum(void) {return wrap_libnum(ll_type);}
L ll_errc[] = {
FMM_HOVF,
FMM_PER,
FMM_INPUT,
FMM_WORK_MEM,
FMM_BOX_SIZE,
FMM_EPS,
FMM_N,
FMM_LIC,
FMM_UNKNOWN,
  0L
};

B* ll_errm[] = {
"** fmm: internal error creating arrays",
"** fmm: performance degraded due to eps",
"** fmm: illegal flag",
"** fmm: work size too small",
"** fmm: size of box generated is zero",
"** fmm: invalid eps",
"** fmm: invalid n",
"** fmm: license error",
"** fmm: unknown error",
  NULL
};

B* ll_export[] = {
  "hi", (B*) op_hi,
  "libnum", (B*) op_libnum,
  "INIT_", (B*) op_INIT_,
  "FINI_", (B*) op_FINI_,
"fmm_kill", (B*) op_fmm_kill,
"fmm_su", (B*) op_fmm_su,
"fmm_calc", (B*) op_fmm_calc,
  "", NULL
};

B opaquename[FRAMEBYTES];
static B FMM_WORK_N[FRAMEBYTES];
static B FMM_MONOCHARGE_N[FRAMEBYTES];
static B FMM_DIPSTR_N[FRAMEBYTES];
static B FMM_POT_N[FRAMEBYTES];
static B FMM_POTX_N[FRAMEBYTES];
static B FMM_POTY_N[FRAMEBYTES];
static B FMM_POTZ_N[FRAMEBYTES];
static B FMM_SOURCES_N[FRAMEBYTES];
static B FMM_DIPVEC_N[FRAMEBYTES];
static B FMM_REFINE_V_N[FRAMEBYTES];
static B FMM_REFINE_X_N[FRAMEBYTES];
static B FMM_REFINE_Y_N[FRAMEBYTES];
static B FMM_REFINE_Z_N[FRAMEBYTES];
static B FMM_INFORM_N[FRAMEBYTES];
static B FMM_OUTPUT_BUF_N[FRAMEBYTES];
static B FMM_OUTPUT_N[FRAMEBYTES];
static B FMM_EPS_CLOSE_N[FRAMEBYTES];
static B FMM_IPREC_N[FRAMEBYTES];
static B FMM_IFLAGTYPE_N[FRAMEBYTES];
static B FMM_MAXWRK_N[FRAMEBYTES];
static B FMM_NAMAX_N[FRAMEBYTES];
static B FMM_IDERIVONOFF_N[FRAMEBYTES];



#define DEBUG_SUM (1)
static doublereal empty_[1];

static D* refine_v = NULL;
static D* refine_x = NULL;
static D* refine_y = NULL;
static D* refine_z = NULL;

static int fmm_close(integer* i_, integer* j_, doublereal* eps,
                     doublereal xinfo1[8], doublereal xinfo2[8],
                     integer iflag[2], doublereal xout[4]) {
  L i = *i_-1;
  L j = *j_-1;

  *(xout++) = xinfo2[3]
    * ((D*) VALUE_PTR(VALUE_PTR(refine_v)+i*FRAMEBYTES))[j];
  *(xout++) = xinfo2[3]
    * ((D*) VALUE_PTR(VALUE_PTR(refine_x)+i*FRAMEBYTES))[j];
  *(xout++) = xinfo2[3]
    * ((D*) VALUE_PTR(VALUE_PTR(refine_y)+i*FRAMEBYTES))[j];
  *(xout++) = xinfo2[3]
    * ((D*) VALUE_PTR(VALUE_PTR(refine_z)+i*FRAMEBYTES))[j];
  
  return 0;
}

static B* lookup_(B* nameframe) {
  return lookup(nameframe, VALUE_PTR(FREEdicts - FRAMEBYTES));
}

  L op_INIT_(void) {
makename("FMM_HANDLE", opaquename);
makename("WORK", FMM_WORK_N);
makename("MONOCHARGE", FMM_MONOCHARGE_N);
makename("DIPSTR", FMM_DIPSTR_N);
makename("POT", FMM_POT_N);
makename("POTX", FMM_POTX_N);
makename("POTY", FMM_POTY_N);
makename("POTZ", FMM_POTZ_N);
makename("SOURCES", FMM_SOURCES_N);
makename("DIPVEC", FMM_DIPVEC_N);
makename("REFINE_V", FMM_REFINE_V_N);
makename("REFINE_X", FMM_REFINE_X_N);
makename("REFINE_Y", FMM_REFINE_Y_N);
makename("REFINE_Z", FMM_REFINE_Z_N);
makename("INFORM", FMM_INFORM_N);
makename("OUTPUT_BUF", FMM_OUTPUT_BUF_N);
makename("OUTPUT", FMM_OUTPUT_N);
makename("EPS_CLOSE", FMM_EPS_CLOSE_N);
makename("IPREC", FMM_IPREC_N);
makename("IFLAGTYPE", FMM_IFLAGTYPE_N);
makename("MAXWRK", FMM_MAXWRK_N);
makename("NAMAX", FMM_NAMAX_N);
makename("IDERIVONOFF", FMM_IDERIVONOFF_N);

  return OK;
}

 L op_FINI_(void) {

  return OK;
}

L op_fmm_kill(void) {

  if (o_1 < FLOORopds) return OPDS_UNF;
  TEST_OPAQUE(o_1);
  FREEopds = o_1;
  KILL_OPAQUE(o1);
  return OK;

}

L op_fmm_su(void) {

  B* frame;
  D* work; L maxwrk;
  D eps_close;
  L iprec;
  L namax;
  L outsize;

  BOOLEAN iderivonoff;
  L iflagtype;
  L sum = 0;

  if (o_7 < FLOORopds) return OPDS_UNF;
  if (CLASS(o_1) != NUM) return OPD_CLA;
  if (! VALUE(o_1, &namax)) return UNDF_VAL;
  if (TAG(o_2) != BOOL) return OPD_TYP;
  iderivonoff = BOOL_VAL(o_2);
  if (CLASS(o_3) != NUM) return OPD_CLA;
  if (! VALUE(o_3, &iflagtype)) return UNDF_VAL;
  if (iflagtype < 0 || iflagtype > 2) return RNG_CHK;
  if (CLASS(o_4) != NUM) return OPD_CLA;
  if (! VALUE(o_4, &iprec)) return UNDF_VAL;
  if (iprec < 0 || iprec > 3) return RNG_CHK;
  if (TAG(o_5) != (NUM | DOUBLETYPE)) return OPD_TYP;
  eps_close = *(D*) NUM_VAL(o_5);
  if (CLASS(o_6) != NUM) return OPD_CLA;
  if (! VALUE(o_6, &maxwrk)) return UNDF_VAL;
  if (CLASS(o_7) != NUM) return OPD_CLA;
  if (! VALUE(o_7, &outsize)) outsize = -1;

  sum += (outsize == -1) ? 0 : DALIGN(FRAMEBYTES + outsize);
  sum += DALIGN(FRAMEBYTES + namax*sizeof(D));
  if (iderivonoff) {
    sum += 3*DALIGN(FRAMEBYTES + namax*sizeof(D));
  }
  if (iflagtype > 0) {
    sum += DALIGN(2*FRAMEBYTES + 4*namax*sizeof(D));
  }
  if (! (iflagtype % 2)) {
    sum += DALIGN(FRAMEBYTES + namax*sizeof(D));
  }
  if (eps_close > 0) {
    sum += 4*DALIGN(FRAMEBYTES*(1 + namax));
    sum += 4*namax*DALIGN(FRAMEBYTES+namax*sizeof(D));
  }
  sum += DALIGN(FRAMEBYTES+3*namax*sizeof(D));
  sum += DALIGN(FRAMEBYTES+5*sizeof(L));
  sum += DALIGN(FRAMEBYTES+maxwrk*sizeof(D));
  
   {
      B initframe[FRAMEBYTES];
      B* procframe = make_opaque_frame(sum, opaquename, 
FMM_WORK_N,
FMM_MONOCHARGE_N,
FMM_DIPSTR_N,
FMM_POT_N,
FMM_POTX_N,
FMM_POTY_N,
FMM_POTZ_N,
FMM_SOURCES_N,
FMM_DIPVEC_N,
FMM_REFINE_V_N,
FMM_REFINE_X_N,
FMM_REFINE_Y_N,
FMM_REFINE_Z_N,
FMM_INFORM_N,
FMM_OUTPUT_BUF_N,
FMM_OUTPUT_N,
FMM_EPS_CLOSE_N,
FMM_IPREC_N,
FMM_IFLAGTYPE_N,
FMM_MAXWRK_N,
FMM_NAMAX_N,
FMM_IDERIVONOFF_N,
NULL);
     if (! procframe) return VM_OVF;

    L off;
    B* buf = VALUE_PTR(OPAQUE_MEM(procframe, buffernameframe));

    ATTR(initframe) = 0;
    off = DALIGN(FRAMEBYTES + maxwrk*sizeof(D));
   if (DEBUG_SUM && (sum -= off) < 0) RETURN_ERROR(FMM_HOVF);
   TAG(initframe)= (ARRAY | DOUBLETYPE); 
   VALUE_PTR(initframe) = buf+FRAMEBYTES;
   ARRAY_SIZE(initframe) = maxwrk;
   OPAQUE_MEM_SET(procframe, FMM_WORK_N, initframe);
   moveframe((initframe), buf);
   buf += off;
    off = DALIGN(FRAMEBYTES + namax*sizeof(D));
   if (DEBUG_SUM && (sum -= off) < 0) RETURN_ERROR(FMM_HOVF);
   TAG(initframe)= (ARRAY | DOUBLETYPE); 
   VALUE_PTR(initframe) = buf+FRAMEBYTES;
   ARRAY_SIZE(initframe) = namax;
   OPAQUE_MEM_SET(procframe, FMM_POT_N, initframe);
   moveframe((initframe), buf);
   buf += off;
    off = DALIGN(FRAMEBYTES + 3*namax*sizeof(D));
   if (DEBUG_SUM && (sum -= off) < 0) RETURN_ERROR(FMM_HOVF);
   TAG(initframe)= (ARRAY | DOUBLETYPE); 
   VALUE_PTR(initframe) = buf+FRAMEBYTES;
   ARRAY_SIZE(initframe) = 3*namax;
   OPAQUE_MEM_SET_NR(procframe, FMM_SOURCES_N, initframe);
   moveframe((initframe), buf);
   buf += off;
    off = DALIGN(FRAMEBYTES + 5*sizeof(L));
   if (DEBUG_SUM && (sum -= off) < 0) RETURN_ERROR(FMM_HOVF);
   TAG(initframe)= (ARRAY | LONGTYPE); 
   VALUE_PTR(initframe) = buf+FRAMEBYTES;
   ARRAY_SIZE(initframe) = 5;
   OPAQUE_MEM_SET(procframe, FMM_INFORM_N, initframe);
   moveframe((initframe), buf);
   buf += off;
    if (outsize != -1) {
      off = DALIGN(FRAMEBYTES + outsize*sizeof(B));
   if (DEBUG_SUM && (sum -= off) < 0) RETURN_ERROR(FMM_HOVF);
   TAG(initframe)= (ARRAY | BYTETYPE); 
   VALUE_PTR(initframe) = buf+FRAMEBYTES;
   ARRAY_SIZE(initframe) = outsize;
   OPAQUE_MEM_SET(procframe, FMM_OUTPUT_BUF_N, initframe);
   moveframe((initframe), buf);
   buf += off;
      if (outsize) {
        ARRAY_SIZE(initframe) = 0;
        OPAQUE_MEM_SET(procframe, FMM_OUTPUT_N, initframe);
      }
    }

    TAG(initframe)= (NUM | LONGTYPE );
    LONG_VAL(initframe) = maxwrk;
    OPAQUE_MEM_SET(procframe, FMM_MAXWRK_N, initframe);

    TAG(initframe)= (NUM | DOUBLETYPE );
    *(D*) NUM_VAL(initframe)= eps_close;
    OPAQUE_MEM_SET(procframe, FMM_EPS_CLOSE_N, initframe);

    TAG(initframe)= (NUM | LONGTYPE );
    *(L*) NUM_VAL(initframe)= iprec;
    OPAQUE_MEM_SET(procframe, FMM_IPREC_N, initframe);

    *(L*) NUM_VAL(initframe)= iflagtype;
    OPAQUE_MEM_SET(procframe, FMM_IFLAGTYPE_N, initframe);

    *(L*) NUM_VAL(initframe)= namax;
    OPAQUE_MEM_SET(procframe, FMM_NAMAX_N, initframe);

    TAG(initframe) = BOOL;
    BOOL_VAL(initframe)= iderivonoff;
    OPAQUE_MEM_SET(procframe, FMM_IDERIVONOFF_N, initframe);

    if (! (iflagtype %2)) {
      off = DALIGN(FRAMEBYTES + namax*sizeof(D));
   if (DEBUG_SUM && (sum -= off) < 0) RETURN_ERROR(FMM_HOVF);
   TAG(initframe)= (ARRAY | DOUBLETYPE); 
   VALUE_PTR(initframe) = buf+FRAMEBYTES;
   ARRAY_SIZE(initframe) = namax;
   OPAQUE_MEM_SET_NR(procframe, FMM_MONOCHARGE_N, initframe);
   moveframe((initframe), buf);
   buf += off;
    }
    if (iflagtype > 0) {
      off = DALIGN(FRAMEBYTES + namax*sizeof(D));
   if (DEBUG_SUM && (sum -= off) < 0) RETURN_ERROR(FMM_HOVF);
   TAG(initframe)= (ARRAY | DOUBLETYPE); 
   VALUE_PTR(initframe) = buf+FRAMEBYTES;
   ARRAY_SIZE(initframe) = namax;
   OPAQUE_MEM_SET_NR(procframe, FMM_DIPSTR_N, initframe);
   moveframe((initframe), buf);
   buf += off;
      off = DALIGN(FRAMEBYTES + 3*namax*sizeof(D));
   if (DEBUG_SUM && (sum -= off) < 0) RETURN_ERROR(FMM_HOVF);
   TAG(initframe)= (ARRAY | DOUBLETYPE); 
   VALUE_PTR(initframe) = buf+FRAMEBYTES;
   ARRAY_SIZE(initframe) = 3*namax;
   OPAQUE_MEM_SET_NR(procframe, FMM_DIPVEC_N, initframe);
   moveframe((initframe), buf);
   buf += off;
    }

    if (iderivonoff) {
      off = DALIGN(FRAMEBYTES + namax*sizeof(D));
   if (DEBUG_SUM && (sum -= off) < 0) RETURN_ERROR(FMM_HOVF);
   TAG(initframe)= (ARRAY | DOUBLETYPE); 
   VALUE_PTR(initframe) = buf+FRAMEBYTES;
   ARRAY_SIZE(initframe) = namax;
   OPAQUE_MEM_SET(procframe, FMM_POTX_N, initframe);
   moveframe((initframe), buf);
   buf += off;
      off = DALIGN(FRAMEBYTES + namax*sizeof(D));
   if (DEBUG_SUM && (sum -= off) < 0) RETURN_ERROR(FMM_HOVF);
   TAG(initframe)= (ARRAY | DOUBLETYPE); 
   VALUE_PTR(initframe) = buf+FRAMEBYTES;
   ARRAY_SIZE(initframe) = namax;
   OPAQUE_MEM_SET(procframe, FMM_POTY_N, initframe);
   moveframe((initframe), buf);
   buf += off;
      off = DALIGN(FRAMEBYTES + namax*sizeof(D));
   if (DEBUG_SUM && (sum -= off) < 0) RETURN_ERROR(FMM_HOVF);
   TAG(initframe)= (ARRAY | DOUBLETYPE); 
   VALUE_PTR(initframe) = buf+FRAMEBYTES;
   ARRAY_SIZE(initframe) = namax;
   OPAQUE_MEM_SET(procframe, FMM_POTZ_N, initframe);
   moveframe((initframe), buf);
   buf += off;
    }

    if (eps_close != 0) {
      {
     B* list = buf; B* listh = buf;
     off = FRAMEBYTES*(1+namax);
     if (DEBUG_SUM && (sum -= off) < 0) RETURN_ERROR(FMM_HOVF);
     TAG(initframe)= LIST;
     VALUE_PTR(initframe) = buf+FRAMEBYTES;
     LIST_CEIL_PTR(initframe) = buf + off;
     OPAQUE_MEM_SET_NR(procframe, FMM_REFINE_V_N, initframe);
     moveframe((initframe), buf);
     buf += off;
     for (list += FRAMEBYTES; list < listh+off; list += FRAMEBYTES) {
       L off2 = DALIGN(FRAMEBYTES + namax*sizeof(D));
       if (DEBUG_SUM && (sum -= off2) < 0) RETURN_ERROR(FMM_HOVF);
       TAG(buf) = (ARRAY | DOUBLETYPE); ATTR(buf) = 0;
       VALUE_PTR(buf) = buf+FRAMEBYTES;
       ARRAY_SIZE(buf) = namax;
       moveframe(buf, list);
       list += FRAMEBYTES;
       buf += off2;
     }
   };
      if (iderivonoff) {
        {
     B* list = buf; B* listh = buf;
     off = FRAMEBYTES*(1+namax);
     if (DEBUG_SUM && (sum -= off) < 0) RETURN_ERROR(FMM_HOVF);
     TAG(initframe)= LIST;
     VALUE_PTR(initframe) = buf+FRAMEBYTES;
     LIST_CEIL_PTR(initframe) = buf + off;
     OPAQUE_MEM_SET_NR(procframe, FMM_REFINE_X_N, initframe);
     moveframe((initframe), buf);
     buf += off;
     for (list += FRAMEBYTES; list < listh+off; list += FRAMEBYTES) {
       L off2 = DALIGN(FRAMEBYTES + namax*sizeof(D));
       if (DEBUG_SUM && (sum -= off2) < 0) RETURN_ERROR(FMM_HOVF);
       TAG(buf) = (ARRAY | DOUBLETYPE); ATTR(buf) = 0;
       VALUE_PTR(buf) = buf+FRAMEBYTES;
       ARRAY_SIZE(buf) = namax;
       moveframe(buf, list);
       list += FRAMEBYTES;
       buf += off2;
     }
   };
        {
     B* list = buf; B* listh = buf;
     off = FRAMEBYTES*(1+namax);
     if (DEBUG_SUM && (sum -= off) < 0) RETURN_ERROR(FMM_HOVF);
     TAG(initframe)= LIST;
     VALUE_PTR(initframe) = buf+FRAMEBYTES;
     LIST_CEIL_PTR(initframe) = buf + off;
     OPAQUE_MEM_SET_NR(procframe, FMM_REFINE_Y_N, initframe);
     moveframe((initframe), buf);
     buf += off;
     for (list += FRAMEBYTES; list < listh+off; list += FRAMEBYTES) {
       L off2 = DALIGN(FRAMEBYTES + namax*sizeof(D));
       if (DEBUG_SUM && (sum -= off2) < 0) RETURN_ERROR(FMM_HOVF);
       TAG(buf) = (ARRAY | DOUBLETYPE); ATTR(buf) = 0;
       VALUE_PTR(buf) = buf+FRAMEBYTES;
       ARRAY_SIZE(buf) = namax;
       moveframe(buf, list);
       list += FRAMEBYTES;
       buf += off2;
     }
   };
        {
     B* list = buf; B* listh = buf;
     off = FRAMEBYTES*(1+namax);
     if (DEBUG_SUM && (sum -= off) < 0) RETURN_ERROR(FMM_HOVF);
     TAG(initframe)= LIST;
     VALUE_PTR(initframe) = buf+FRAMEBYTES;
     LIST_CEIL_PTR(initframe) = buf + off;
     OPAQUE_MEM_SET_NR(procframe, FMM_REFINE_Z_N, initframe);
     moveframe((initframe), buf);
     buf += off;
     for (list += FRAMEBYTES; list < listh+off; list += FRAMEBYTES) {
       L off2 = DALIGN(FRAMEBYTES + namax*sizeof(D));
       if (DEBUG_SUM && (sum -= off2) < 0) RETURN_ERROR(FMM_HOVF);
       TAG(buf) = (ARRAY | DOUBLETYPE); ATTR(buf) = 0;
       VALUE_PTR(buf) = buf+FRAMEBYTES;
       ARRAY_SIZE(buf) = namax;
       moveframe(buf, list);
       list += FRAMEBYTES;
       buf += off2;
     }
   };
      }
    }

    if (DEBUG_SUM && sum) RETURN_ERROR(FMM_HOVF);

     moveframe(procframe, o_7);
   }

    FREEopds = o_6;
    return OK;

}

L op_fmm_calc(void) {

  D* work; L maxwrk;
  D eps_close;
  D *monocharge = empty_, *dipstr = empty_;
  D *pot, *potx = empty_, *poty = empty_, *potz = empty_;
  D *sources, *dipvec = empty_;
  L iprec;

  BOOLEAN iderivonoff;
  L iflagtype;
  L* inform;
  L namax;
  L outsize = -1; B* outbuf = NULL;

  integer ier;
  integer inform_[5];
  L i;

  if (o_1 < FLOORopds) return OPDS_UNF;
  TEST_OPAQUE(o_1);

  work = (D*) FMM_WORK(o_1);
  maxwrk = FMM_MAXWRK(o_1);
  eps_close = FMM_EPS_CLOSE(o_1);
  sources = (D*) FMM_SOURCES(o_1);
  iprec = FMM_IPREC(o_1);
  iderivonoff = FMM_IDERIVONOFF(o_1);
  iflagtype = FMM_IFLAGTYPE(o_1);
  inform = (L*)FMM_INFORM(o_1);
  namax = FMM_NAMAX(o_1);
  pot = (D*)FMM_POT(o_1);

  refine_v = empty_;
  refine_x = empty_;
  refine_y = empty_;
  refine_z = empty_;

  if (TAG(OPAQUE_MEM(o_1, FMM_OUTPUT_BUF_N))
       != NULLOBJ)
  outbuf = (B*) FMM_OUTPUT_BUF(o_1);
if (TAG(OPAQUE_MEM(o_1, FMM_MONOCHARGE_N))
       != NULLOBJ)
  monocharge = (D*) FMM_MONOCHARGE(o_1);
if (TAG(OPAQUE_MEM(o_1, FMM_DIPSTR_N))
       != NULLOBJ)
  dipstr = (D*) FMM_DIPSTR(o_1);
if (TAG(OPAQUE_MEM(o_1, FMM_POTX_N))
       != NULLOBJ)
  potx = (D*) FMM_POTX(o_1);
if (TAG(OPAQUE_MEM(o_1, FMM_POTY_N))
       != NULLOBJ)
  poty = (D*) FMM_POTY(o_1);
if (TAG(OPAQUE_MEM(o_1, FMM_POTZ_N))
       != NULLOBJ)
  potz = (D*) FMM_POTZ(o_1);
if (TAG(OPAQUE_MEM(o_1, FMM_DIPVEC_N))
       != NULLOBJ)
  dipvec = (D*) FMM_DIPVEC(o_1);
if (TAG(OPAQUE_MEM(o_1, FMM_REFINE_V_N))
       != NULLOBJ)
  refine_v = (D*) FMM_REFINE_V(o_1);
if (TAG(OPAQUE_MEM(o_1, FMM_REFINE_X_N))
       != NULLOBJ)
  refine_x = (D*) FMM_REFINE_X(o_1);
if (TAG(OPAQUE_MEM(o_1, FMM_REFINE_Y_N))
       != NULLOBJ)
  refine_y = (D*) FMM_REFINE_Y(o_1);
if (TAG(OPAQUE_MEM(o_1, FMM_REFINE_Z_N))
       != NULLOBJ)
  refine_z = (D*) FMM_REFINE_Z(o_1);

  if (outbuf) {
    outsize = FMM_OUTPUT_BUF_SIZE(o_1);
    if (outsize)
      prin_set_output_indirect(outbuf, outsize);
    else
      prin_set_output_none();
  }
  else
    prin_set_output_direct();

  ier = fmmpart3d(f2c(sources, 3),
                  namax,
                  monocharge,
                  dipstr,
                  f2c(dipvec, 3),
                  pot, potx, poty, potz,
                  work,
                  maxwrk,
                  iprec,
                  iflagtype,
                  iderivonoff ? 1 : 0,
                  fmm_close,
                  eps_close,
                  inform_);

  for (i = 0; i < 5; ++i) inform[i] = inform_[i];

  switch (ier) {
    case 0: break;
    case 1: RETURN_ERROR(FMM_PER);
    case 2: RETURN_ERROR(FMM_INPUT);
    case 3: RETURN_ERROR(FMM_WORK_MEM);
    case 4: RETURN_ERROR(FMM_BOX_SIZE);
    case 5: RETURN_ERROR(FMM_EPS);
    case 6: RETURN_ERROR(FMM_N);
    case 7: RETURN_ERROR(FMM_LIC);
    default: RETURN_ERROR(FMM_UNKNOWN);
  };

  if (outbuf && outsize) {
    FMM_OUTPUT_SIZE(o_1) = ind_prin_used_get();
  }

  FREEopds = o_1;
  return OK;

}


