#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>

#include "fmmbase-util.h"
#include "fmmbase-io.h"

static char* ind_prin_arr = NULL;
static size_t ind_prin_len = 0;
static char* ind_prin_base = NULL;
static size_t ind_prin_used = 0;

int ind_prin_used_get(void) {return ind_prin_used;}

typedef enum {fmm_io_none, fmm_io_direct, fmm_io_indirect} fmm_io_mode;

static fmm_io_mode mode = fmm_io_direct;

void prin_set_output_direct(void) {
  ind_prin_base = NULL;
  ind_prin_used = 0;
  mode = fmm_io_direct;
}

void prin_set_output_indirect(char* arr, size_t len) {
  ind_prin_arr = arr;
  ind_prin_len = len;
  ind_prin_base = arr;
  ind_prin_used = 0;
  mode = fmm_io_indirect;
}

void prin_set_output_none(void) {
  ind_prin_base = NULL;
  ind_prin_used = 0;
  mode = fmm_io_none;
}

static void non_prinf_(const char* mes, const integer* ia, integer n, 
                       ftnlen mes_len) {
  return;
}

static void non_prin2_(const char* mes, const doublereal* a2, integer n, 
                      ftnlen mes_len) {
  return;
}

static int ind_printf(const char* format, ...)
  __attribute ((__format__ (__printf__, 1, 2)));

static int ind_printf(const char* format, ...) {
  size_t s;
  va_list ap;
  if (! ind_prin_len) return 0;

  va_start(ap, format);
  if ((s = vsnprintf(ind_prin_arr, ind_prin_len, format, ap)) > ind_prin_len) 
    s = ind_prin_len;
  va_end(ap);

  ind_prin_len -= s;
  ind_prin_arr += s;
  ind_prin_used += s;

  return ind_prin_len;
}

static const char* strnchr(const char* s, size_t count, int c) {
  const char* s_ = s + count;
  for (; s < s_ && *s; s++) {
    if (*s == c) return s;
  }
  return NULL;
}

static int ind_mespr(const char* mes, ftnlen mes_len) {
  int imes_len = mes_len > 10000 ? 10000 : mes_len;
  const char* ast = strnchr(mes, imes_len, '*');
  const char* imes = mes;
  int len = ast ? (ast - mes) : imes_len;

  for (; len > 0; len -= 80, imes += 80)
    if (! ind_printf(" %.*s%s", len < 80 ? len:80, imes, len < 80 ? "":"\n"))
      return 0;

  return 1;
}

static void ind_prinf_(const char* mes, const integer* ia, integer n, 
                      ftnlen mes_len) {
  int i;
  if (! ind_mespr(mes, mes_len)) return;
  
  for (i = 1; i <= n; i++)
    if (! ind_printf(" %7d%s", f1(ia, i), !(i%10) ? "\n" : ""))
      return;
  if (i%10) ind_printf("\n");
}

static void ind_prin2_(const char* mes, const doublereal* a2, integer n, 
                       ftnlen mes_len) {
  int i;

  if (! ind_mespr(mes, mes_len)) return;
  for (i = 1; i <= n; i++)
    if (! ind_printf("  %11.5e%s", f1(a2, i), !(i%6) ? "\n" : ""))
      return;
  if (i%6) ind_printf("\n");
}

static void dir_mespr(const char* mes, ftnlen mes_len) {
  int imes_len = mes_len > 10000 ? 10000 : mes_len;
  const char* imes = mes;
  const char* ast = strnchr(imes, imes_len, '*');
  int len = ast ? (ast - mes) : imes_len;

  for (; len > 0; len -= 80, imes += 80)
    printf(" %.*s%s", len < 80 ? len : 80, imes, len < 80 ? "" : "\n");
}

static void dir_prinf_(const char* mes, const integer* ia, integer n, 
                       ftnlen mes_len) {
  int i;

  dir_mespr(mes, mes_len);
  
  for (i = 1; i <= n; i++)
    printf(" %7d%s", f1(ia, i), !(i%10) ? "\n" : "");
  if (i%10) printf("\n");
}
 
static void dir_prin2_(const char* mes, const doublereal* a2, integer n, 
                       ftnlen mes_len) {
  int i;
  
  dir_mespr(mes, mes_len);
  for (i = 1; i <= n; i++)
    printf("  %11.5e%s", f1(a2, i), !(i%6) ? "\n" : "");
  if (i%6) printf("\n");
} 

int prinf_(const char* mes, const integer* ia, const integer* n, 
            ftnlen mes_len) {
  switch (mode) {
    case fmm_io_direct:   dir_prinf_(mes, ia, *n, mes_len); break;
    case fmm_io_indirect: ind_prinf_(mes, ia, *n, mes_len); break;
    case fmm_io_none:     non_prinf_(mes, ia, *n, mes_len); break;
  }
  return 0;
}

int prin2_(const char* mes, const doublereal* a2, const integer* n,
           ftnlen mes_len) {
  switch (mode) {
    case fmm_io_direct:   dir_prin2_(mes, a2, *n, mes_len); break;
    case fmm_io_indirect: ind_prin2_(mes, a2, *n, mes_len); break;
    case fmm_io_none:     non_prin2_(mes, a2, *n, mes_len); break;
  }
  return 0;
}

