#include "test-common.h"
#define DEBUG_DLARAN (0)

#if DEBUG_DLARAN
#define out(var)                                \
  {                                             \
    int j;                                      \
    for (j = 0; j < 4; j++) {                   \
      fprintf(stderr, "i[%d]=%d,", j, iseed[j]);  \
    }                                             \
    doublereal r = dlaran(iseed);               \
    fprintf(stderr, " : %s %e, ", #var, r);        \
    var r;                                      \
  }
#else
#define out(var) var dlaran(iseed)
#endif

int main(int argc, char* argv[]) {
  integer ier;
  integer inform[5];
  integer iprec;
  integer iflagtype;
  integer iderivonoff;
  integer nsources;
  integer i;
  doublereal epsclose;

  integer iseed[4], itarg;

  printf("Test # 1001\n");
  f1(iseed, 1) = 1000;
  f1(iseed, 2) = 1000;
  f1(iseed, 3) = 1000;
  f1(iseed, 4) = 0001;

  printf("\n=== Generating particle system\n\n");
  nsources = 4500;
  for (i = 1; i <= nsources; i++) {
    if (DEBUG_DLARAN) fprintf(stderr, "#%d: ", i);
    out(f2(sources, 1, i) =);
    out(f2(sources, 2, i) =);
    out(f2(sources, 3, i) =);
    out(f1(monocharge, i) = 1 - 2*);
    out(f1(dipstr, i) = );
    out(f2(dipvec, 1, i) = 1 - 2*);
    out(f2(dipvec, 2, i) = 1 - 2*);
    out(f2(dipvec, 3, i) = 1 - 2*);
    if (DEBUG_DLARAN) fprintf(stderr, "\n");
  }

  iprec = 3;
  iflagtype = 2;
  iderivonoff = 1;
  epsclose = 0;
  
  printf("\n=== Calling FFM\n\n");
  ier = fmmpart3d(sources, nsources, monocharge, dipstr,
                  dipvec, pot, potx, poty, potz, work,
                  maxwrk, iprec, iflagtype, iderivonoff,
                  close1, epsclose, inform);
  if (ier != 0) 
    printf("WARNING, error code returned. ier = %d\n", ier);
  
  i = 2;
  printf("pot(%d) = %e\npotx(%d) = %e\npoty(%d) = %e\npotz(%d) = %e\n", 
         i, f1(pot, i), i, f1(potx, i), i, f1(poty, i), i, f1(potz, i));
  printf("\n=== Direct evaluation\n\n");
  if (nsources < 5000) {
    doublereal error;
    direct_eval(sources, nsources, monocharge, dipstr, dipvec,
                pot2, potx2, poty2, potz2, work, maxwrk,
                iflagtype, iderivonoff, 1, nsources);
    printf("=== Comparing the results\n\n");
    printf("Mean square error in potential = %e\n",
            compare_meansq(pot, pot2, nsources));
    for (i = 1; i <= nsources; i++) {
      f1(work, i) = f1(potx, i);
      f1(work, nsources   + i) = f1(poty, i);
      f1(work, 2*nsources + i) = f1(potz, i);
      f1(work, 3*nsources + i) = f1(potx2, i);
      f1(work, 4*nsources + i) = f1(poty2, i);
      f1(work, 5*nsources + i) = f1(potz2, i);
    }
    printf("Mean square error in field = %e\n",
           compare_meansq(&f1(work, 1), &f1(work, 3*nsources+1), 3*nsources));
  }
  else {
    itarg = 1 + (nsources - 2) * dlaran(iseed);
    printf("We check only one, random, point.\n\n");
    
    direct_eval(sources, nsources, monocharge, dipstr,
                dipvec, pot2, potx2, poty2, potz2, work, maxwrk,
                iflagtype, iderivonoff, itarg, itarg);
    printf("At itarg = %d\n", itarg);
    printf("pot = %e, %e\n", f1(pot, itarg), f1(pot2, itarg));
    printf("potx = %e, %e\n", f1(potx, itarg), f1(potx2, itarg));
    printf("poty = %e, %e\n", f1(poty, itarg), f1(poty2, itarg));
    printf("potz = %e, %e\n", f1(potz, itarg), f1(potz2, itarg));
    printf("rel error in pot = %e\n",
           abs(f1(pot, itarg) - f1(pot2, itarg))/abs(f1(pot2, itarg)));
    printf("rel error in field = %e\n",
           3*(abs(f1(potx, itarg) - f1(potx2, itarg))
              + abs(f1(poty, itarg) - f1(poty2, itarg))
              + abs(f1(potz, itarg) - f1(potz2, itarg)))
           / (abs(f1(potx2, itarg) 
                  + abs(f1(poty2, itarg)) 
                  + abs(f1(potz2, itarg)))));
  }
  
  return 0;
}


